///////////////////////////////////////////////////////////////////
// This software is copyrighted by Optical Communications Group. //
// University of Valladolid (UVa)                                //
//                                                               //
// WRON simulator                                                //
// has been developed by Optical Communications Group.           //
// University of Valladolid (UVa)                                //
// Contact e-mail: rduran@tel.uva.es,natalia.fernandez@tel.uva.es //
//                                                               //
// Escuela Técnica Superior de Ingenieros de Telecomunicación    //
// Campus Miguel Delibes                                         //
// Universidad de Valladolid                                     //
// Paseo de Belén, 15                                            //
// 47011 Valladolid, Spain                                       //
//                                                               //
// This file is part of CHRON simulator,                         //
// an optical network simulation tool                            //
// based on OMNeT++/OMNEST discrete-event simulation framework.  //
//                                                               //
// Authoring:                                                    //
// Natalia Fernández Sordo (natalia.fernandez@tel.uva.es)        //
// Ramón J. Durán Barroso (rduran@tel.uva.es)                    //
// Ignacio de Miguel Jiménez (ignacio.miguel@tel.uva.es)         //
///////////////////////////////////////////////////////////////////

#ifndef CTOPOLOGYEXTENSION_H_
#define CTOPOLOGYEXTENSION_H_

#include <string>
#include <vector>
#include <omnetpp.h>
#include "Estado.h"
#include "cDelayLengthChannel.h"

using namespace std;

NAMESPACE_BEGIN

class cPar;

// not all compilers define INFINITY (gcc does)
#ifndef INFINITY
#define INFINITY HUGE_VAL
#endif

/**
 * Routing support. The cTopologyExtension class is an extension of cTopology class.
 *
 * A cTopology object stores an abstract representation of the network
 * in graph form:
 * <UL>
 *   <LI> each cTopology node corresponds to a getModule(simple or compound), and
 *   <LI> each cTopology edge corresponds to a link or series of connecting links.
 * </UL>
 *
 * You can specify which modules (either simple or compound) you want to
 * include in the graph. The graph will include all connections among the
 * selected modules. In the graph, all nodes are at the same level, there's
 * no submodule nesting. Connections which span across compound module
 * boundaries are also represented as one graph edge. Graph edges are directed,
 * just as module gates are.
 *
 *
 * @ingroup SimSupport
 * @see cTopologyExtension::NodeExtension, cTopologyExtension::LinkExtension, cTopologyExtension::LinkInExtension, cTopologyExtension::LinkOutExtension
 */
class SIM_API cTopologyExtension : public cTopology
{
  public:
    class LinkExtension;
    class LinkInExtension;
    class LinkOutExtension;

    /**
     * Supporting class for cTopologyExtension, represents a node in the graph. This class includes more attributes than cTopology::Node
     */
    class SIM_API NodeExtension : public Node
    {
        friend class cTopologyExtension;

      private:
        int module_id;
        double wgt;
        bool enabl;
        int numberEDFAS;

        int num_in_links;
        LinkExtension **in_links; //This type is modified to be implemented here
        int num_out_links;
        LinkExtension *out_links; //This type is modified to be implemented here

        // variables used by the shortest-path algorithms
        bool known;
        double dist;

        LinkExtension *out_path; //This type is modified to be implemented here
        int num_paths_m; //Number of paths which are the shortest paths to a destination
        std::vector<NodeExtension *> node_paths_m; //A NodeExtension vector which includes Nodes in the shortest-path algorithm
        std::vector<int> num_links_m; //
        std::vector<LinkExtension *>array_links_path; //A LinkExtension vector which includes Nodes in the shortest-path algorithm
        Estado *state_module;

      public:
        /** @name Node attributes: weight, enabled state, correspondence to modules. */
        //@{

        /**
         * Returns the ID of the network module to which this node corresponds.
         */
        int getModuleId() const  {return module_id;}

        /**
         * Returns the pointer to the network module to which this node corresponds.
         */
        cModule *getModule() const  {return simulation.getModule(module_id);}

        /**
         * Returns the weight of this node. Weight is used with the
         * weighted shortest path finder methods of cTopology.
         */
        double getWeight() const  {return wgt;}

        /**
         * Sets the weight of this node. Weight is used with the
         * weighted shortest path finder methods of cTopology.
         */
        void setWeight(double d)  {wgt=d;}



        int getEDFAS() const {return numberEDFAS;}
        void setEDFAS(int edfas) {numberEDFAS=edfas;}

        /**
         * Returns true of this node is enabled. This has significance
         * with the shortest path finder methods of cTopology.
         */
        bool isEnabled() const  {return enabl;}

        /**
         * Enable this node. This has significance with the shortest path
         * finder methods of cTopology.
         */
        void enable()  {enabl=true;}

        /**
         * Disable this node. This has significance with the shortest path
         * finder methods of cTopology.
         */
        void disable()  {enabl=false;}
        //@}

        /** @name Node connectivity. */
        //@{

        /**
         * Returns the number of incoming links to this graph node.
         */
        int getNumInLinks() const  {return num_in_links;}

        /**
         * Returns ith incoming link of graph node.
         */
        LinkInExtension *getLinkIn(int i);

        /**
         * Returns the number of outgoing links from this graph node.
         */
        int getNumOutLinks() const  {return num_out_links;}

        /**
         * Returns ith outgoing link of graph node.
         */
        LinkOutExtension *getLinkOut(int i);
        //@}

        /** @name Result of shortest path extraction. */
        //@{

        /**
         * Returns the distance of this node to the target node.
         */
        double getDistanceToTarget() const  {return dist;}

        /**
         * Returns the number of shortest paths towards the target node.
         * (There may be several paths with the same length.)
         */
        int getNumPaths() const {return out_path?1:0;}

        /**
         * Returns the next link in the ith shortest paths towards the
         * target node. (There may be several paths with the same
         * length.)
         */

         LinkOutExtension *getPath(int);
        //@}

         /**
		 * Returns the number of shortest paths towards the target node.
         * (There may be several paths with the same length.)
         */
        int numPaths_m() const                 {return num_paths_m;}

        /**
         * Returns the next node in the ith shortest paths towards the
         * target node. (There may be several paths with the same
         * length.)
         */
        NodeExtension *paths_m(int);




    };


    /**
     * Supporting class for cTopologyExtension, represents a link in the graph.
     * This is a linkExtension to be used in the cTopology class
     */
    class SIM_API LinkExtension: public Link
    {
        friend class cTopologyExtension;

      protected:
        NodeExtension *src_node;
        int src_gate;
        NodeExtension *dest_node;
        int dest_gate;
        double wgt;
        bool enabl;
        int numberEDFAS;


      public:
        /**
         * Returns the weight of this link. Weight is used with the
         * weighted shortest path finder methods of cTopology.
         */
        double getWeight() const  {return wgt;}

        /**
         * Sets the weight of this link. Weight is used with the
         * weighted shortest path finder methods of cTopology.
         */
        void setWeight(double d)  {wgt=d;}

        int getEDFAS() const {return numberEDFAS;}
        void setEDFAS(int edfas) {numberEDFAS=edfas;}

        /**
         * Returns true of this link is enabled. This has significance
         * with the shortest path finder methods of cTopology.
         */
        bool isEnabled() const  {return enabl;}

        /**
         * Enables this link. This has significance with the shortest path
         * finder methods of cTopology.
         */
        void enable()  {enabl=true;}

        /**
         * Disables this link. This has significance with the shortest path
         * finder methods of cTopology.
         */
        void disable()  {enabl=false;}
    };


    /**
     * Supporting class for cTopology.
     *
     * While navigating the graph stored in a cTopology, Node's methods return
     * LinkIn and LinkOut objects, which are 'aliases' to Link objects.
     * LinkIn and LinkOut provide convenience functions that return the
     * 'local' and 'remote' end of the connection when traversing the topology.
     */
    class SIM_API LinkInExtension : public LinkExtension
    {
      public:
        /**
         * Returns the node at the remote end of this connection.
         *
         * Note: There's no corresponding localNode() method: the local node of
         * this connection is the Node object whose method returned
         * this LinkIn object.
         */
        NodeExtension *getRemoteNode() const  {return src_node;}

        /**
         * Returns the gate ID at the remote end of this connection.
         */
        int getRemoteGateId() const  {return src_gate;}

        /**
         * Returns the gate ID at the local end of this connection.
         */
        int getLocalGateId() const  {return dest_gate;}

        /**
         * Returns the gate at the remote end of this connection.
         */
        cGate *getRemoteGate() const  {return src_node->getModule()->gate(src_gate);}

        /**
         * Returns the gate at the local end of this connection.
         */
        cGate *getLocalGate() const  {return dest_node->getModule()->gate(dest_gate);}
    };


    /**
     * Supporting class for cTopology.
     *
     * While navigating the graph stored in a cTopology, Node's methods return
     * LinkIn and LinkOut objects, which are 'aliases' to Link objects.
     * LinkIn and LinkOut provide convenience functions that return the
     * 'local' and 'remote' end of the connection when traversing the topology.
     */
    class SIM_API LinkOutExtension : public LinkExtension
    {
      public:
        /**
         * Returns the node at the remote end of this connection.
         *
         * Note: There's no corresponding localNode() method: the local node of
         * this connection is the Node object whose method returned
         * this LinkIn object.
         */
        NodeExtension *getRemoteNode() const  {return dest_node;}

        /**
         * Returns the gate ID at the remote end of this connection.
         */
        int getRemoteGateId() const  {return dest_gate;}

        /**
         * Returns the gate ID at the local end of this connection.
         */
        int getLocalGateId() const  {return src_gate;}

        /**
         * Returns the gate at the remote end of this connection.
         */
        cGate *getRemoteGate() const  {return dest_node->getModule()->gate(dest_gate);}

        /**
         * Returns the gate at the local end of this connection.
         */
        cGate *getLocalGate() const  {return src_node->getModule()->gate(src_gate);}
    };

    /**
     * Base class for selector objects used in extract...() methods of cTopology.
     * Redefine the matches() method to return whether the given module
     * should be included in the extracted topology or not.
     */
    class SIM_API PredicateExtension
    {
      public:
        virtual ~PredicateExtension() {}
        virtual bool matches(cModule *module) = 0;
    };

  protected:
    int num_nodes;
    NodeExtension *nodev;
    NodeExtension *target;

  public:
    /** @name Constructors, destructor, assignment */
    //@{

    /**
     * Constructor.
     */
    explicit cTopologyExtension(const char *name=NULL);

    /**
     * Copy constructor.
     */
    cTopologyExtension(const cTopologyExtension& topo);

    /**
     * Destructor.
     */
    virtual ~cTopologyExtension();

    /**
     * Assignment operator. The name member doesn't get copied; see cNamedObject's operator=() for more details.
     */
    cTopologyExtension& operator=(const cTopologyExtension& topo);
    //@}

    /** @name Redefined cObject member functions. */
    //@{

    /**
     * Creates and returns an exact copy of this object.
     * See cObject for more details.
     */
    virtual cTopologyExtension *dup() const  {return new cTopologyExtension(*this);}

    /**
     * Produces a one-line description of the object's contents.
     * See cObject for more details.
     */
    virtual std::string info() const;

    /**
     * Serializes the object into an MPI send buffer.
     * Used by the simulation kernel for parallel execution.
     * See cObject for more details.
     */
    virtual void parsimPack(cCommBuffer *buffer);

    /**
     * Deserializes the object from an MPI receive buffer
     * Used by the simulation kernel for parallel execution.
     * See cObject for more details.
     */
    virtual void parsimUnpack(cCommBuffer *buffer);
    //@}

    /** @name Extracting the topology from a network.
     *
     * extract...() functions build topology from the model.
     * User can select which modules to include. All connections
     * between those modules will be in the topology. Connections can
     * cross compound module boundaries.
     */
    //@{

    /**
     * Extracts model topology by a user-defined criteria. Includes into the graph
     * modules for which the passed selfunc() returns nonzero. The userdata
     * parameter may take any value you like, and it is passed back to selfunc()
     * in its second argument.
     */
    void extractFromNetwork(bool (*selfunc)(cModule *,void *), void *userdata=NULL);

    /**
     * The type safe, object-oriented equivalent of extractFromNetwork(selfunc, userdata).
     */
    void extractFromNetwork(PredicateExtension *predicate);

    /**
     * Extracts model topology by module full path. All modules whole getFullPath()
     * matches one of the patterns in given string vector will get included.
     * The patterns may contain wilcards in the same syntax as in ini files.
     *
     * An example:
     *
     * <tt>topo.extractByModulePath(cStringTokenizer("**.host[*] **.router*").asVector());</tt>
     */
    void extractByModulePath(const std::vector<std::string>& fullPathPatterns);

    /**
     * Extracts model topology by the fully qualified NED type name of the
     * modules. All modules whose getNedTypeName() is listed in the given string
     * vector will get included.
     *
     * Note: If you have all class names as a single, space-separated string,
     * you can use cStringTokenizer to turn it into a string vector:
     *
     * <tt>topo.extractByNedTypeName(cStringTokenizer("some.package.Host other.package.Router").asVector());</tt>
     */
    void extractByNedTypeName(const std::vector<std::string>& nedTypeNames);

    /**
     * Extracts model topology by a module property. All modules get included
     * that have a property with the given name and the given value
     * (more precisely, the first value of its default key being the specified
     * value). If value is NULL, the property's value may be anything except
     * "false" (i.e. the first value of the default key may not be "false").
     *
     * For example, <tt>topo.extractByProperty("node");</tt> would extract
     * all modules that contain the <tt>@node</tt> property, like the following
     * one:
     *
     * <pre>
     * module X {
     *     @node;
     * }
     * </pre>
     *
     */
    void extractByProperty(const char *propertyName, const char *value=NULL);

    /**
     * Extracts model topology by a module parameter. All modules get included
     * that have a parameter with the given name, and the parameter's str()
     * method returns the paramValue string. If paramValue is NULL, only the
     * parameter's existence is checked but not its value.
     */
    void extractByParameter(const char *paramName, const char *paramValue=NULL);

    /**
     * Deletes the topology stored in the object.
     */
    void clear();
    //@}

    /** @name Functions to examine topology by hand.
     *
     * Users also need to rely on Node and Link member functions
     * to explore the graph stored in the object.
     */
    //@{

    /**
     * Returns the number of nodes in the graph.
     */
    int getNumNodes() const  {return num_nodes;}

    /**
     * Returns pointer to the ith node in the graph. Node's methods
     * can be used to further examine the node's connectivity, etc.
     */
    NodeExtension *getNode(int i);

    /**
     * Returns the graph node which corresponds to the given module in the
     * network. If no graph node corresponds to the module, the method returns
     * NULL. This method assumes that the topology corresponds to the
     * network, that is, it was probably created with one of the
     * extract...() functions.
     */
    NodeExtension *getNodeFor(cModule *mod);
    //@}

    /** @name Algorithms to find shortest paths. */
    //@{

    /**
     * Apply the Dijkstra algorithm to find a shortest paths to the given
     * graph node. The paths found can be extracted via NodeExtension's methods.
     */
    void calculateUnweightedSingleShortestPathsTo(NodeExtension *target);

    /**
	* Apply the Dijkstra algorithm to find all shortest paths to the given
	* graph node. The paths found can be extracted via NodeExtension's methods.
	*/
    void calculateUnweightedMultipleShortestPathsTo(NodeExtension *target);

   /**
	* Apply the Dijkstra algorithm to find all weighted shortest paths to the given
	* graph node. The paths found can be extracted via Node's methods.
	*/
	void calculateWeightedMultipleShortestPathsTo(NodeExtension *target);

    /**
     * Extracts model topology by a user-defined criteria. Includes into the graph
     * modules for which the passed selfunc() returns nonzero. The userdata
     * parameter may take any value you like, and it is passed back to selfunc()
     * in its second argument.
     */
	void extractFromWeightedNetwork(bool (*selfunc)(cModule *,void *), void *userdata=NULL);

	/**
	 * Extracts model topology by module type (classnames). Includes into
	 * the graph all modules whose className() is one of the strings
	 * listed as arguments. The argument list must be terminated by NULL.*/

	  void extractWeigths(const std::vector<std::string>& nedTypeNames);

    /**
     * Returns the node that was passed to the most recently called
     * shortest path finding function.
     */
    NodeExtension *getTargetNode() const {return target;}
    //@}



	void NoWavelengthsConsideration();


	void virtualExtractByNedTypeName(const std::vector<std::string>& nedTypeNames, vector < vector < int> >,int);

	void virtualExtractFromNetwork(vector < vector < int> >, int ,bool (*selfunc)(cModule *,void *), void *userdata=NULL);


};

NAMESPACE_END



#endif /* CTOPOLOGYEXTENSION_H_ */
